# @cloudbase/functions-framework 使用文档

`@cloudbase/functions-framework` 是一个函数式代码运行框架，本文档主要说明如何使用 `@cloudbase/functions-framework` 框架。

`@cloudbase/functions-framework` 可将如下结构的函数代码：

```js
exports.main = function(event, context) {
  return 'Hello, World!'
}
```

运行起来，并通过 HTTP 请求触发云函数执行：

```sh
curl http://url
# 输出：Hello, World!
```

将函数代码打包后，可以进一步部署在 Docker 等虚拟化环境中。

当前框架支持的 Node.js 版本为 `>=18.0.0`。

## 特性

- 单实例多并发的请求处理模式，更贴近一般 Web 服务的运行方式，支持 SSE、Websocket 等长连接
- 相较于云函数（SCF）提供更高的灵活性，无定制运行时，无需关注函数实例的生命周期
- 原生面向 HTTP，无需理解复杂的自定义事件概念
- 支持加载多函数，支持函数间路由，可直接在不同函数间复用代码
- 提供面向 Web 服务的丰富配置项，如超时时间、并发限制、优雅停止等
- 面向用户的多层次、结构化日志输出，支持自定义日志轮转策略、日志打印内容
- 本地与线上环境同构，支持热重启模式，易于在本地调试

## 安装与运行

执行以下命令安装：

```sh
npm i -g @cloudbase/functions-framework
```

安装后执行以下输出查看是否安装成功：

```sh
tcb-ff -h
```

### 直接通过命令行运行

有如下目录结构：

```sh
/workspace
├── index.js
├── package.json
```

在 `index.js` 中有如下代码：

```js
exports.main = function (event, context) {
  const { httpContext } = context
  const { url, httpMethod } = httpContext
  return `[${httpMethod}][${url}] Hello world!`
}
```

在 `workspace` 目录下执行如下命令，启动监听 `3000` 端口的服务：

```sh
tcb-ff
```

未指定参数时，框架读取的默认函数文件为 `index.js`，默认函数名称为 `main`，默认监听端口为 `3000`。

### 注册函数

有如下目录结构：

```sh
/workspace
├── fn.js
└── package.json
```

在 `fn.js` 中引入`@cloudbase/functions-framework`：

```js
// 引入注册函数的方法
import { event } from '@cloudbase/functions-framework'

// 注册一个命名为 `main` 的函数
event("main", async function (event, context) {
  const { httpContext } = context
  const { url, httpMethod } = httpContext
  return `[${httpMethod}][${url}] Hello world!`
})
```

在 `package.json` 中，使用 `@cloudbase/functions-framework` 作为依赖，并配置启动命令：

```json
{
  "name": "example",
  "version": "1.0.0",
  "type": "module",
  "main": "fn.js",
  "scripts": {
    "start": "npx @cloudbase/functions-framework"
  },
  "dependencies": {
    "@cloudbase/functions-framework": "^0.0.9"
  }
}
```

使用 `npm start` 启动服务。

### 进一步封装

如果希望自行解析参数，可在解析后将解析好的配置作为参数，手动启动服务：

```js
import { startFrameworkService } from '@cloudbase/functions-framework'

// 导出函数供外部调用
export function yourFunction() {
  // 解析参数
  const options = {
    // ...
  }
  // 启动服务
  startFrameworkService(options)
}
```

**如何在封装框架的情况下支持`-w`/`--watch`（热重启）功能？**
由于框架的热重启依赖于启动文件，如果重新封装参数解析，可能导致热重启不起作用。可以参考 `bin/tcb-ff.js` 中的热重启实现，适配自己的 CLI 支持即可。

## 参数说明

可通过 `tcb-ff -h` 打印帮助信息查看所有参数。

| 命令行参数            | 环境变量              | 默认值                              | 示例                                              | 说明                                                                                                                                                                         |
| --------------------- | --------------------- | ----------------------------------- | ------------------------------------------------- | ---------------------------------------------------------------------------------------------------------------------------------------------------------------------------- |
| --port                | PORT                  | 3000                                | --port=3000                                       | 监听的端口                                                                                                                                                                   |
| --target              | FUNCTION_TARGET       | main                                | --target=hello                                    | 目标函数名称                                                                                                                                                                 |
| --source              | FUNCTION_SOURCE       | index.js or index.cjs or index.mjs  | --source=main.mjs                                 | 目标函数文件名                                                                                                                                                               |
| --signatureType       | SIGNATURE_TYPE        | event                               | --signatureType=event                             | 函数格式，当前仅支持`event`                                                                                                                                                  |
| --gracefulShutdown    | GRACEFUL_SHUTDOWN     | `NODE_ENV` 不为 `production` 时开启 | --gracefulShutdown=false                          | 是否启用服务的优雅停止                                                                                                                                                       |
| /                     | KILL_TIMEOUT          | 30000                               | KILL_TIMEOUT=60000                                | 优雅停止时，等待请求处理完成的超时时间                                                                                                                                       |
| --timeout             | TIMEOUT               | 0                                   | --timeout=5000                                    | Node.js Server 超时时间，见 <https://nodejs.org/api/http.html#servertimeout>                                                                                                 |
| --keepAliveTimeout    | KEEP_ALIVE_TIMEOUT    | 65000                               | --keepAliveTimeout=10000                          | Node.js Server keep-alive 超时时间，见 <https://nodejs.org/api/http.html#server-keepalivetimeout>                                                                            |
| --captureStack        | CAPTURE_STACK         | `NODE_ENV` 不为 `production` 时开启 | --captureStack=true                               | 请求出错时，是否返回错误堆栈                                                                                                                                                 |
| --concurrencyLimit    | CONCURRENCY_LIMIT     | 1000                                | --concurrencyLimit=100                            | 并发请求数限制，0 为不限制                                                                                                                                                   |
| --noExit              | NO_EXIT               | false                               | --noExit=true                                     | 是否禁用 `process.exit()` 避免进程意外退出                                                                                                                                   |
| --logHeaderBody       | LOG_HEADER_BODY       | false                               | --logHeaderBody=true                              | 是否在请求日志中记录请求头、请求体、响应头、响应体，如记录，将会限制 body 大小为 2KB                                                                                         |
| --logEventContext     | LOG_EVENT_CONTEXT     | true                                | --logEventContext=false                           | 是否在请求日志中记录入参 `event` 和 `context`                                                                                                                                |
| --logDirname          | LOG_DIRNAME           | ./logs                              | --logDirname=/var/logs                            | 日志文件目录                                                                                                                                                                 |
| --maxLogFileSize      | MAX_LOG_FILE_SIZE     | `1024 * 1024 * 128`(128m)           | --maxLogFileSize=`2097152`                        | 单个日志文件最大字节数                                                                                                                                                       |
| --maxLogFiles         | MAX_LOG_FILES         | 4                                   | --maxLogFiles=5                                   | 最多同时存在的日志文件数量                                                                                                                                                   |
| -w, --watch           | /                     | false                               | -w                                                | 是否启用热重启模式，如启用，将会在函数代码文件变更时自动重启服务                                                                                                             |
| --interceptOutput     | INTERCEPT_OUTPUT      | false                               | --interceptOutput=true                            | 是否拦截`console.log`等标准输出，若拦截，会将标准输出的内容记录到日志文件中                                                                                                  |
| --extendedContextKey  | EXTENDED_CONTEXT_KEY  | `''`                                | --extendedContextKey=X-Functions-Extended-Context | 指定获取扩展 `context.extendedContext` 的头部字段                                                                                                                            |
| --enableCors          | ENABLE_CORS           | false                               | --enableCors=true                                 | 是否启用 CORS，启用后可配置放通的域名，默认放通 `localhost` 和 `127.0.0.1`                                                                                                   |
| --allowedOrigins      | ALLOWED_ORIGINS       | `localhost,127.0.0.1`               | --allowedOrigins="example.com,*.abc.com"          | 需放通 CORS 的域名，默认放通 `localhost` 和 `127.0.0.1`。支持通配符，如 `*.example.com` 可匹配 `www.example.com` 和 `auth.example.com`。该配置项仅在启用`enableCors`时生效。 |
| --functionsConfigFile | FUNCTIONS_CONFIG_FILE | cloudbase-functions.json            | --functionsConfigFile=myconfig.json               | 多函数路由配置文件名                                                                                                                                                         |
| --dry-run             | /                     | false                               | --dry-run                                         | 是否启用验证模式，如启用，将不会启动服务，仅检查函数是否可以正常加载                                                                                                         |

## 运行 TypeScript 代码

借助 [`ts-node`](https://www.npmjs.com/package/ts-node)，`tcb-ff` 可以直接加载 `TypeScript` 函数代码，使本地开发体验更为流畅。

要直接运行 TypeScript 代码，首先需要安装 `ts-node`：

```sh
npm i -D ts-node
```

然后通过 `NODE_OPTIONS` 环境变量指定 `ts-node` 的加载方式：

```sh
# 直接加载 TypeScript 函数
NODE_OPTIONS="--require ts-node/register/transpile-only" tcb-ff
# 直接加载 TypeScript 函数，并启用本地热更新
NODE_OPTIONS="--require ts-node/register/transpile-only" tcb-ff -w
# 直接加载 TypeScript 函数，并启用调试模式，启动后可支持调试 TypeScript 代码
NODE_OPTIONS="--require ts-node/register/transpile-only --inspect=localhost:9229" tcb-ff
```

> 注意：`ts-node` 仅在开发环境中使用，生产环境请先编译 TypeScript 代码后再运行。

## 使用 VSCode 调试函数代码（JavaScript/TypeScript）

通过 vscode 可调试 `JavaScript/TypeScript` 编写的函数代码，需创建 `.vscode/launch.json` 配置文件，并配置如下参考内容：

> 注意：以下配置可根据注释描述进行调整

```json
{
  // Use IntelliSense to learn about possible attributes.
  // Hover to view descriptions of existing attributes.
  // For more information, visit: https://go.microsoft.com/fwlink/?linkid=830387
  "version": "0.2.0",
  "configurations": [
    {
      "type": "node",
      "request": "launch",
      "name": "launch-tcb-ff-global",
      // 注意：请确保全局安装了 @cloudbase/functions-framework 模块
      // npm install -g @cloudbase/functions-framework
      // 因不同环境下全局安装路径可能不同，需要手动替换
      //  Windows 系统可通过 `where tcb-ff` 命令查看全局安装路径
      //  MacOS/Linux 系统下可通过 `which tcb-ff` 命令查看全局安装路径
      "program": "please-replace-this-with-the-path-to-your-global-tcb-ff",
      // 直接启动 src 目录下的 ts 函数代码，如需调试编译后的 js 代码，可将 --source 参数替换为编译后的目录
      "args": ["--source src"],
      "env": {
          // 默认支持调试 TS 代码，如不需要可以移除
          "NODE_OPTIONS": "--require ts-node/register/transpile-only"
      },
      "skipFiles": [
        "<node_internals>/**"
      ],
      "outFiles": [
        "!**/node_modules/**"
      ]
    },
    {
      "type": "node",
      "request": "launch",
      "name": "launch-tcb-ff-local",
      // 注意：请确保安装在当前项目下安装了 @cloudbase/functions-framework 模块
      // npm install @cloudbase/functions-framework
      "program": "${workspaceFolder}/node_modules/.bin/tcb-ff",
      // 直接启动 src 目录下的 ts 函数代码，如需调试编译后的 js 代码，可将 --source 参数替换为编译后的目录
      "args": ["--source src"],
      "env": {
          // 默认支持调试 TS 代码，如不需要可以移除
          "NODE_OPTIONS": "--require ts-node/register/transpile-only"
      },
      "skipFiles": [
        "<node_internals>/**"
      ],
      "outFiles": [
        "!**/node_modules/**"
      ]
    },
    {
      // 注意：在启动调试进程前，需确保已通过 `NODE_OPTIONS='--inspect=localhost:9229' tcb-ff .` 等调试命令启动了服务
      "address": "localhost",
      "port": 9229, // 配置端口为 9229
      "localRoot": "${workspaceFolder}",
      // 采用 Attach Node.js 进程的方式调试
      "name": "attach-to-remote-9229",
      "request": "attach",
      "skipFiles": [
        "<node_internals>/**"
      ],
      "type": "node"
    }
  ]
}
```

以上配置文件中包含了两种类型（`launch`/`attach`）共三个调试配置：

1. [推荐] vscode `launch` 启动 node 进程进行调试
   1. `launch-tcb-ff-global`：使用全局安装的 `tcb-ff` 运行函数代码
   2. `launch-tcb-ff-local`：使用本地安装的 `tcb-ff` 运行函数代码
  
   通过 `launch` 方式调试 `ts` 代码，实际存在两种模式：

     1. 利用 `ts-node` 加载 TypeScript 代码能力，`tcb-ff` 直接运行 `src` 目录下的 `ts` 函数代码进行调试，该方式不需要提前编译代码。
     2. `tcb-ff` 运行 `编译后的目录` 下的 `js` 函数代码进行调试，该方式需要先运行 `npx tsc -w` 编译 `ts` 代码。

2. vscode `attach` 方式调试
   1. `attach-to-remote-9229`：Attach 到远程 9229 端口的 Node.js 进程

    使用该方式调试，需要在终端独立执行启动云函数服务命令，调试步骤如下：

    首先，在终端手动执行如下命令：

    ```sh
    # 指定调试服务器端口为 9229
    NODE_OPTIONS="--inspect=localhost:9229" tcb-ff
    ```

    然后，在 vscode 中，通过快捷键（`F5`）启动调试进程（或在 `Run And Debug` 界面选择 `attach-to-remote-9229` 启动调试），即可开始进行断点调试。

## 构建镜像

### 构建`@cloudbase/functions-framework`镜像

在 `docker/framework` 目录下，提供了构建`@cloudbase/functions-framework`镜像的 Dockerfile，参考其 README 即可构建指定版本的框架镜像。

### 构建函数镜像

使用指定版本的`@cloudbase/functions-framework`镜像构建函数镜像时，可将`docker/Dockerfile`复制到函数代码根目录下，并参考 README 构建任意函数镜像。

## 局限

1. 虽然函数框架提供了许多配置，但不支持自定义 HTTP 服务，也不支持更改提供 HTTP 服务的框架。
2. 函数框架并不和云函数（SCF）兼容，因此无法直接将云函数代码迁移到函数框架中。
3. 函数框架无法做到请求粒度的调度，最小调度单位为函数实例，单个函数实例支持多并发。
4. 由于一个函数实例在逻辑上对应着一个服务而非一个请求，如果划分过多的函数，可能导致服务数量过多。
